*-------------------------------------------------------------------------
*   Arithmetic overflow checking for Atari ST Icon   (rover.asm)
*-------------------------------------------------------------------------
*   Stack arrangement for a C function call
*       (reference version 3.03 Lattice C Manual, page 242)
*-------------------------------------------------------------------------
*
*                ========   SP (a7) ->  |-------------------| Low
*                expression evaluation  |    temporaries    |
*                                       |-------------------|
*                addressed as negative  |  auto data items  |
*                      offsets from FP  |     .  .  .       |
*                ========   FP (a6) ->  |-------------------|
*                called routine pushes  | caller's saved FP |
*                                       |-------------------|
*                pushed by jsr          |  return address   |
*                                       |-------------------|
*                first argument is      |    arguments      |
*                a6 plus 8 bytes        |     .  .  .       |
*                old SP value   ----->  |-------------------| High
*
*-------------------------------------------------------------------------
        text                         control section for instructions
        xdef    ckmul
        xdef    ckadd
        xdef    cksub
        xref    exit
        xref    runerr
*-------------------------------------------------------------------------
*   Do integer multiply and test for overflow  (ckmul)
*-------------------------------------------------------------------------
*   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
ckmul   equ     *
        link    a6,#0                push frame pointer & allocate no locals
        move.l  d2,a0                save data registers in address registers
        move.l  d3,a1                    to preserve across calls
        moveq   #0,d3                initialize sign flag to zero
*
*  Get first operand, make positive, and note sign
*
        move.l  8(a6),d1             get first operand  (aa bb)
        beq     zeroret              answer is zero if op is zero
        bpl     posone
        addq    #1,d3                toggle sign flag
        neg.l   d1                   complement operand to positive
        bmi     ovrflw               overflow if sign bit is set
posone  equ     *
*
*  Get second operand, make positive, and note sign
*
        move.l  12(a6),d2            get second operand (cc dd)
        beq     zeroret              answer is zero if op is zero
        bpl     postwo
        addq    #1,d3                toggle sign flag
        neg.l   d2                   complement operand to positive
        bmi     ovrflw               overflow if sign bit is set
postwo  equ     *
*
*  Begin by multiplying lower halves of each operand
*
        move.l  d1,d0                make a copy of operand
        mulu    d2,d0                d0 = (bb * dd)
        bmi     ovrflw               overflow if sign bit is set
*
*  Check upper halves of operands; either (aa) or (cc) must be zero
*
        swap    d1                   make upper halves available
        swap    d2
        tst.w   d1                   if (aa) is not the zero
        bne     opone
        tst.w   d2                   if (cc) is not the zero
        bne     optwo
        bra     sgnfix               d0 has answer if both (aa) and (cc) zero
*
*  Case (aa) not zero
*
opone   equ     *
        tst.w   d2
        bne     ovrflw               overflow if both (aa) and (cc) not zero
        swap    d2
        bra     combine
*
*  Case (cc) not zero
*
optwo   equ     *                    (aa) already tested as zero
        swap    d1
*       bra     combine
*
*  Add either (aa * dd) or (cc * bb) to (bb * dd) for unsigned answer
*             ---------    ---------    ---------
*
combine equ     *
        mulu    d2,d1                d1 = (aa * dd) or (cc * bb)
        cmpi.l  #$7fff,d1            overflow indicated if too large
        bgt     ovrflw                   for a 16-bit shift left
        swap    d1                   equivalent to 16-bit shift left
        add.l   d1,d0                combine the two partial answers
        bvs     ovrflw               addition may result in overflow
*
*  Fix the sign if necessary--cannot cause overflow now
*
sgnfix  equ     *
        ror     #1,d3
        bcc     nofix                no sign adjustment needed
        neg.l   d0                   otherwise answer is negative
nofix   equ     *
*
*  Clean up for return
*
done    move.l  a0,d2                restore data registers
        move.l  a1,d3
        unlk    a6                   restore frame pointer
        rts                          return to calling program

*
*  Zero value returned
*
zeroret moveq   #0,d0
        bra     done
*
*  Overflow indicated
*
ovrflw  equ     *
        moveq   #0,d0
        move.l  d0,-(a7)             push arguments 0 and 203
        move.l  #-203,-(a7)              for runerr()
        jsr     runerr               call runerr()
*   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
*  Following instructions not needed because runerr() never returns         *
*                                                                           *
        addq.l  #8,a7                remove runerr() args from stack        *
        move.l  #0,-(a7)             put argument on stack for exit()       *
        jsr     exit                 and exit the program                   *
*       bra     done                                                        *
*   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
*-------------------------------------------------------------------------
*   Do integer addition and test for overflow  (ckadd)
*-------------------------------------------------------------------------
*   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
ckadd   equ     *
        link    a6,#0                push frame pointer & allocate no locals
        move.l  8(a6),d0             get first operand
        add.l   12(a6),d0            add second operand
        bvs     ovrflw               check for overflow
        unlk    a6                   restore frame pointer
        rts                          return to calling program
*-------------------------------------------------------------------------
*   Do integer subtraction and test for overflow  (cksub)
*-------------------------------------------------------------------------
*   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
cksub   equ     *
        link    a6,#0                push frame pointer & allocate no locals
        move.l  8(a6),d0             get first operand
        sub.l   12(a6),d0            subtract second operand
        bvs     ovrflw               check for overflow
        unlk    a6                   restore frame pointer
        rts                          return to calling program
        end
*-------------------------------------------------------------------------
