         /* A version of EXTCALL for accessing/modifying Rexx
            (or CLIST) variables from Icon.  See REXX.ICN for
            intended usage. */

#include "config.h"
#include "rt.h"
#include "rproto.h"

#ifdef ExternalFunctions

static struct descrip rxretv;
static int rxrc;
static int rxrestart, rxvlen, rxvnlen;
static char rxlastvn[256];
               /* last variable name retrieved */
static char rxlastvl[MaxCvtLen];
               /* value of last variable retrieved */

static dptr extnext(),extvar(), extset(), extdrop();
static int rxerr(), makemsg();

#include <exec.h>

#define compare(a,b) (StrLen(a) != sizeof(b) - 1 || \
                      memcmp(StrLoc(a), (b), sizeof(b) - 1))
               /* efficient compare of string to constant */

/*
 * extcall - stub procedure for external call interface.
 */
dptr extcall(dargv, argc, ip)
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];

   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;
      return dargv;
   }
   if (!compare(*dargv, "rexxvar")) return extvar(dargv, argc, ip);
   if (!compare(*dargv, "rexxset")) return extset(dargv, argc, ip);
   if (!compare(*dargv, "rexxdrop")) return extdrop(dargv, argc, ip);
   if (!compare(*dargv, "rexxnext")) return extnext(dargv, argc, ip);
   *ip = 216;                   /* no external function to find */
   return (dptr)NULL;
   }

static dptr extset(dargv, argc, ip)  /* set a Rexx variable */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf1[MaxCvtLen];
   char sbuf2[MaxCvtLen];

   if (argc < 3) {      /* too few arguments */
      *ip = 103;
      return NULL;
   }
   ++dargv;
   if (cvstr(dargv, sbuf1) == CvtFail) {
      *ip = 103;        /* non-string argument */
      return dargv;
   }
   if (cvstr(dargv+1, sbuf2) == CvtFail) {
      *ip = 103;
      return dargv+1;
   }
   rxrestart = 0;       /* FIRST/NEXT cycle must resume */
   rxrc = execshv(SHV_SET, StrLoc(*dargv), StrLen(*dargv),
                  StrLoc(dargv[1]), StrLen(dargv[1]), 0);
   if (rxrc == SHV_SUCCESS)
      return &nulldesc;
   else rxrc = rxerr(rxrc, ip);      /* if failure, return error msg */
   if (rxrc)
      return &rxretv;
   else return (dptr) NULL;
   }

static dptr extvar(dargv, argc, ip)  /* return value of variable */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];
   int rxlen;

   if (argc < 2) {      /* too few arguments */
      *ip = 103;
      return NULL;
   }
   ++dargv;
   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;        /* non-string argument */
      return dargv;
   }

         /* If this call is to get the value of the variable whose
            name was last returned by extnext, the value is already
            in rxlastvl. */

   if (rxrestart == 1 && rxvnlen == StrLen(*dargv) &&
       !memcmp(StrLoc(*dargv), rxlastvn, rxvnlen)) {
      if (rxvlen <= MaxCvtLen) {     /* if the value wasn't trunc'd */
         if (strreq(rxvlen) == Error) {
            *ip = 0;    /* unable to allocate string space */
            return 0;
         }
         StrLen(rxretv) = rxvlen+1;
         StrLoc(rxretv) = alcstr("=", 1);  /* flag value as value,
                                              not message */
         (void) alcstr(rxlastvl, rxvlen);
         return &rxretv;
      } else rxrestart = 2;          /* if value was truncated, it   */
                                     /* must be retrieved again,     */
                                     /* which will force restart     */
                                     /* on the next call to extnext  */
   } else rxrestart = 0;
   rxrc = execshv(SHV_FETCH, StrLoc(*dargv), StrLen(*dargv), sbuf,
                  MaxCvtLen, &rxlen);     /* fetch the variable */
   if (rxrc == SHV_NOT_FOUND) return NULL;
                                     /* fail if not found */
   if (rxrc == SHV_TRUNC_VAL || rxrc == SHV_SUCCESS) {
      if (strreq(rxlen) == Error) {
         *ip = 0;
         return 0;
      }
      if (rxrc == SHV_TRUNC_VAL) {   /* if the value was cut off, */
                                     /* fetch all of it */
         StrLoc(rxretv) = alcstr(NULL, rxlen+1);
         *StrLoc(rxretv) = '=';      /* value, not message */
         execshv(SHV_FETCH, StrLoc(*dargv), StrLen(*dargv),
                 StrLoc(rxretv)+1, rxlen, &rxlen);
      } else {
         StrLoc(rxretv) = alcstr("=", 1); /* value, not message */
         (void) alcstr(sbuf, rxlen);
      }
      StrLen(rxretv) = rxlen+1;
      return &rxretv;
   } else rxrc = rxerr(rxrc, ip);    /* generate message for
                                        unexpected stuff */
   if (rxrc)
      return &rxretv;
   else return (dptr) NULL;
   }

static dptr extdrop(dargv, argc, ip) /* drop a Rexx variable */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];

   if (argc < 2) {   /* too few args */
      *ip = 103;
      return NULL;
   }
   ++dargv;
   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;     /* arg not string */
      return dargv;
   }
   rxrestart = 0;    /* show FIRST-NEXT broken */
   rxrc = execshv(SHV_DROP, StrLoc(*dargv), StrLen(*dargv), 0, 0, 0);
   if (rxrc == SHV_NOT_FOUND || rxrc == SHV_SUCCESS)
      return &nulldesc;    /* return &null if it's dropped */
   else rxrc = rxerr(rxrc, ip);  /* else return failure message */
   if (rxrc)
      return &rxretv;
   else return (dptr) NULL;
   }

static dptr extnext(dargv, argc, ip) /* return next variable name */
dptr dargv;
int argc;
int *ip;
   {
   int rxrc;
   char rxlastvnsv[256];

         /* Rexx will return to us, for each variable, its name
            and value.  This call returns the name, and extvar is
            gimmicked to avoid calling Rexx again for the call to
            get the value which is anticipated (unless the value
            was truncated, in which case we must start over). */

   if (rxrestart == 0) {   /* if we must start over */
      rxrestart = 1;
      rxrc = execshv(SHV_FIRST, rxlastvn, 256, rxlastvl, MaxCvtLen,
                     &rxvlen);       /* ask for the first one */
   } else if (rxrestart == 2) {      /* if we must start over &
                                        resume in the middle */
      strcpy(rxlastvnsv, rxlastvn);  /* save where we were */
      rxrc = execshv(SHV_FIRST, rxlastvn, 256, rxlastvl, MaxCvtLen,
                     &rxvlen);       /* start over */
   } else {
again:
      rxrc = execshv(SHV_NEXT, rxlastvn, 256, rxlastvl, MaxCvtLen,
                     &rxvlen);       /* ask for the next one */
   }
   if (rxrc == SHV_LAST_VAR) {       /* fail after last one */
      rxrestart = 0;
      return NULL;
   } else if(rxrc == SHV_SUCCESS || rxrc == SHV_TRUNC_VAL) {
      if (rxrestart == 2) {          /* decide whether to stop skipping */
         if(strcmp(rxlastvn, rxlastvnsv) == 0) rxrestart = 1;
         goto again;
      }
      rxvnlen = strlen(rxlastvn);
      if (strreq(rxvnlen) == Error) {
         *ip = 0;
         return 0;
      }
      StrLoc(rxretv) = alcstr("=", 1);
      (void) alcstr(rxlastvn, rxvnlen);  /* return variable name */
      StrLen(rxretv) = rxvnlen+1;
      return &rxretv;
   } else rxrc = rxerr(rxrc);        /* or message in ugly cases */
   if (rxrc)
      return &rxretv;
   else return (dptr) NULL;
   }

static int makemsg(text, ip)    /* turn a message into an Icon string */
char *text;
int *ip;
   {
   int len;

   len = strlen(text);
   if (strreq(len) == Error) {
      *ip = 0;
      return 0;
   }
   StrLoc(rxretv) = alcstr(text, len);
   StrLen(rxretv) = len;
   return 1;
   }

static int rxerr(rc,ip)    /* generate a message for strange Rexx
                              conditions */
int rc;
int *ip;
   {
      switch(rc) {
case SHV_NO_SUBCOM:
         rc = makemsg("No EXEC is active", ip);
         break;
case SHV_NO_MEM:
         rc = makemsg("Out of memory", ip);
         break;
case SHV_INVALID_VAR:
         rc = makemsg("Invalid variable name", ip);
         break;
case SHV_NOT_SUPPORTED:
         rc = makemsg("Not supported by operating system", ip);
         break;
default:
         rc = makemsg("Rexx interface failed", ip);
         break;
      }
      return rc;
   }
#else                                   /* ExternalFunctions */
static char x;                  /* prevent empty module */
#endif                                  /* ExternalFunctions */
